<?php

namespace Bitrix\Modules\Sale\Handler\Paysystem\Paykeeper;

use Bitrix\Sale\BusinessValue;
use Bitrix\Sale\Order;
use Bitrix\Sale\PaySystem\Service;
use Bitrix\Main\Web\HttpClient;
use Bitrix\Main\Event;
use Bitrix\Main\EventResult;
use Bitrix\Main\EventManager;
use Bitrix\Main\Loader;
use Bitrix\Main\Context;

use Bitrix\Sale\Internals\PaymentTable;
use Bitrix\Sale\Internals\PayableItemTable;

// Registering the event handler for the 'OnSaleStatusOrder' event in the 'sale' module.
EventManager::getInstance()->addEventHandler(
    'sale',
    'OnSaleStatusOrder',
    [__NAMESPACE__ . '\PaykeeperEventHandler','OnSaleStatusOrderChange']
);

define('LOG_FILE', realpath(dirname(__FILE__)) . "/logs/paykeeper.log");

/**
 * Class PaykeeperEventHandler
 * Handles events and actions related to Paykeeper integration.
 */
class PaykeeperEventHandler {

    const log_file = LOG_FILE;
    private static $logging = false;

    /**
     * Event handler for the OnSaleStatusOrderChange event.
     *
     * This method is triggered when the order status changes. It handles the final receipt generation
     * and interaction with the Paykeeper API if certain conditions are met.
     *
     * @param int $orderId The ID of the order whose status has changed.
     * @return EventResult Returns an EventResult object, indicating the success of the operation.
     */
    static function OnSaleStatusOrderChange($orderId)
    {
        $order = Order::load($orderId);
        if (!$order->isPaid()) {
            return self::getSuccessEventResult();
        }

        $paySystemId = $order->getField('PAY_SYSTEM_ID');
        $personTypeId = $order->getField('PERSON_TYPE_ID');

        $paykeeper_hold_funds_enable = self::getBusVal('PAYKEEPER_HOLD_FUNDS_ENABLE', $order);
        $paykeeper_print_final_receipt = self::getBusVal('PAYKEEPER_PRINT_FINAL_RECEIPT', $order);
        $paykeeper_lk_login = self::getBusVal('PAYKEEPER_LK_LOGIN', $order);
        $paykeeper_lk_password = self::getBusVal('PAYKEEPER_LK_PASSWORD', $order);
        self::$logging = (self::getBusVal('PAYKEEPER_LOGGING', $order) === 'Y');

        if (!$paykeeper_lk_login
            || !$paykeeper_lk_password
            || ($paykeeper_print_final_receipt !== 'Y' && $paykeeper_hold_funds_enable !== 'Y'))
        {
            return self::getSuccessEventResult();
        }

        $paykeeper_url = self::getBusVal('PAYKEEPER_FORM_URL', $order);
        if (!$paykeeper_url) {
            return self::getSuccessEventResult();
        }
        if ($pos = strrpos($paykeeper_url, '/create') ?: strrpos($paykeeper_url, '/order')) {
            $paykeeper_url = substr($paykeeper_url, 0, $pos);
        }

        $paykeeper_hold_debit_after_status = self::getBusVal('PAYKEEPER_HOLD_DEBIT_AFTER_STATUS', $order);
        $paykeeper_hold_cancel_after_status = self::getBusVal('PAYKEEPER_HOLD_CANCEL_AFTER_STATUS', $order);
        $paykeeper_print_after_status = self::getBusVal('PAYKEEPER_PRINT_AFTER_STATUS', $order);

        $order_status_id = $order->getField('STATUS_ID');

        $paykeeper_is_receipt = $paykeeper_print_final_receipt === 'Y'
            && $paykeeper_print_after_status == $order_status_id;
        $paykeeper_is_hold_debit = $paykeeper_hold_funds_enable === 'Y'
            && $paykeeper_hold_debit_after_status == $order_status_id;
        $paykeeper_is_hold_cancel = $paykeeper_hold_funds_enable === 'Y'
            && $paykeeper_hold_cancel_after_status == $order_status_id;

        if (!$paykeeper_is_receipt && !$paykeeper_is_hold_debit && !$paykeeper_is_hold_cancel) {
            return self::getSuccessEventResult();
        }

        // Получаем ID платежа РК
        $paymentCollection = $order->getPaymentCollection();
        $paymentId = 0;
        foreach ($paymentCollection as $payment) {
            if ($payment->getField('PAY_SYSTEM_ID') != $paySystemId) {
                continue;
            }
            $paymentId = preg_replace('/\D/', '', $payment->getField('PS_STATUS_MESSAGE'));
        }

        if (!$paymentId) {
            return self::getSuccessEventResult();
        }

        /* Авторизация в Paykeeper */
        $paykeeper_nttp_client = new HttpClient();
        $basic_auth = base64_encode("$paykeeper_lk_login:$paykeeper_lk_password");
        $paykeeper_nttp_client->setHeader('Content-Type','application/x-www-form-urlencoded');
        $paykeeper_nttp_client->setHeader('Authorization','Basic '.$basic_auth);
        $result_paykeeper_get = $paykeeper_nttp_client->get("$paykeeper_url/info/settings/token/");
        $response = json_decode($result_paykeeper_get, true);
        if (self::$logging) {
            $log_params = [
                'TITLE' => 'EVENT',
                'URL' => "$paykeeper_url/info/settings/token/",
                'METHOD' => 'GET',
                'RESPONSE' => $response
            ];
            self::paykeeperLogger($log_params);
        }
        $error_flag = false;
        $params_for_log = array();

        $eventName = '';
        $logEventName = '';
        $infoLog = '';
        if ($paykeeper_is_receipt) {
            $eventName = 'onPrintCheckPaykeeper';
            $logEventName = 'ГЕНЕРАЦИЯ ЧЕКА';
        } else if($paykeeper_is_hold_debit) {
            $eventName = 'onHoldDebitPaykeeper';
            $logEventName = 'АВТОРИЗАЦИЯ СРЕДСТВ';
        } else if($paykeeper_is_hold_cancel) {
            $eventName = 'onHoldCancelPaykeeper';
            $logEventName = 'ОТМЕНА АВТОРИЗАЦИИ';
        }

        /* В ответе должен быть token, иначе ошибка */
        if (!isset($response['token'])) {
            $error_flag = true;
            $infoLog = 'Заказ '.$orderId.'. Ошибка авторизации!';
            self::triggerPaykeeperEventError($infoLog, $orderId, $eventName);
        } else {
            $token = $response['token'];

            // Event Print Receipt
            if ($paykeeper_is_receipt) {
                $response = $paykeeper_nttp_client->post(
                    "$paykeeper_url/change/payment/post-sale-receipt/",
                    ['id' => $paymentId, 'token' => $token]
                );
                $response = json_decode($response, true);

                // Создаем новую группу свойств и свойство для чека окончательного расчета, если их нет
                $propertyCode = 'PAYKEEPER_RECEIPT_ID';
                $propertyName = 'Номер запроса на генерацию чека в PK';
                self::createOrderPropertyForAllPersonTypes($propertyName, $propertyCode);

                $receipt_id = 0;
                /* В ответе должен быть receipt_id */
                if (isset($response['receipt_id'])) {
                    $receipt_id = $response['receipt_id'];
                    $infoLog = 'Заказ '.$orderId.'. ID чека ' . $receipt_id;

                    // Записываем значение номера запроса печати чека
                    /** @var \Bitrix\Sale\PropertyValueCollection $propertyCollection */
                    $propertyCollection = $order->getPropertyCollection();
                    $property = $propertyCollection->getItemByOrderPropertyCode($propertyCode);

                    if($property && $property->getValue() != $receipt_id) {
                        $property->setField('VALUE', $receipt_id);
                        $order->save();
                    }
                } elseif (isset($response['error'])) {
                    $error_flag = true;
                    $infoLog = 'Заказ '.$orderId.'. Ошибка: '.$response['error'];
                } elseif (isset($response['msg'])) {
                    $error_flag = true;
                    $infoLog = 'Заказ '.$orderId.'. Ошибка: '.$response['msg'];
                } else {
                    $error_flag = true;
                    $infoLog = 'Заказ '.$orderId.'. Неизвестная ошибка!';
                }

                $params_for_log = array(
                    'SEVERITY' => 'INFO',
                    'AUDIT_TYPE_ID' => 'Печать чека окончательного расчета PayKeeper',
                    'MODULE_ID' => 'sale',
                    'ITEM_ID' => 'API запрос /change/payment/post-sale-receipt/',
                    'DESCRIPTION' => $infoLog,
                );
            }

            // Event Hold Debit
            if ($paykeeper_is_hold_debit) {
                $response = $paykeeper_nttp_client->post(
                    "$paykeeper_url/change/payment/capture/",
                    ['id' => $paymentId, 'token' => $token]
                );
                $response = json_decode($response, true);
                if (self::$logging) {
                    $log_params = [
                        'TITLE' => 'PAYMENT_CAPTURE',
                        'URL' => "$paykeeper_url/change/payment/capture/",
                        'METHOD' => 'POST',
                        'DATA' => ['id' => $paymentId, 'token' => 'XXXX'],
                        'RESPONSE' => $response
                    ];
                    self::paykeeperLogger($log_params);
                }

                $response = isset($response[0]) ? $response[0] : $response;

                if (isset($response['result']) && $response['result'] == 'success') {
                    $infoLog = 'Заказ ' . $orderId . '. Списание средств!';
                } else if (isset($response['error'])) {
                    $error_flag = true;
                    $infoLog = 'Заказ '.$orderId.'. Ошибка: '.$response['error'];
                } else if (isset($response['msg'])) {
                    $error_flag = true;
                    $infoLog = 'Заказ '.$orderId.'. Ошибка: '.$response['msg'];
                } else {
                    $error_flag = true;
                    $infoLog = 'Заказ '.$orderId.'. Неизвестная ошибка!';
                }

                $params_for_log = array(
                    'SEVERITY' => 'INFO',
                    'AUDIT_TYPE_ID' => 'Списание средств по ранее проведённой авторизации PayKeeper',
                    'MODULE_ID' => 'sale',
                    'ITEM_ID' => 'API запрос /change/payment/capture/',
                    'DESCRIPTION' => $infoLog,
                );
            }

            // Event Hold Cancel
            if ($paykeeper_is_hold_cancel) {
                $response = $paykeeper_nttp_client->post(
                    "$paykeeper_url/change/payment/reverse/",
                    [
                        'id' => $paymentId,
                        'amount' => self::getBusVal('PAYKEEPER_FORM_URL', $order),
                        'partial' => false,
                        'token' => $token
                    ]
                );
                $response = json_decode($response, true);
                if (self::$logging) {
                    $log_params = [
                        'TITLE' => 'PAYMENT_REVERSE',
                        'URL' => "$paykeeper_url/change/payment/reverse/",
                        'METHOD' => 'POST',
                        'DATA' => [
                            'id' => $paymentId,
                            'amount' => self::getBusVal('PAYKEEPER_FORM_URL', $order),
                            'partial' => false,
                            'token' => 'XXXX'
                        ],
                        'RESPONSE' => $response
                    ];
                    self::paykeeperLogger($log_params);
                }

                $response = isset($response[0]) ? $response[0] : $response;

                if (isset($response['result']) && $response['result'] == 'success') {
                    $infoLog = 'Заказ '.$orderId.'. Отмена авторизации!';
                } else if (isset($response['error'])) {
                    $error_flag = true;
                    $infoLog = 'Заказ '.$orderId.'. Ошибка: '.$response['error'];
                } else if (isset($response['msg'])) {
                    $error_flag = true;
                    $infoLog = 'Заказ '.$orderId.'. Ошибка: '.$response['msg'];
                } else {
                    $error_flag = true;
                    $infoLog = 'Заказ '.$orderId.'. Неизвестная ошибка!';
                }

                $params_for_log = array(
                    'SEVERITY' => 'INFO',
                    'AUDIT_TYPE_ID' => 'Отмена авторизации средств PayKeeper',
                    'MODULE_ID' => 'sale',
                    'ITEM_ID' => 'API запрос /change/payment/reverse/',
                    'DESCRIPTION' => $infoLog,
                );
            }
        }

        // Добавляем запись в журнал событий об API-запросе
        if ($params_for_log) {
            \CEventLog::Add($params_for_log);
        }

        // Добавляем всплывающее уведомление в административной панели
        if ($error_flag) {
            self::triggerPaykeeperEventError($infoLog, $orderId, $eventName);

            \CAdminNotify::Add(array(
                'TAG' => "PAYKEEPER_$orderId",
                'MESSAGE' => "$logEventName PAYKEEPER. $infoLog",
                'NOTIFY_TYPE' => 'E'
            ));
        }

        return self::getSuccessEventResult();
    }

    /**
     * Creates an order property for all person types.
     *
     * This method checks if a property with a specific code exists for each person type in the system.
     * If not, it creates the property and assigns it to a property group.
     *
     * @param string $propertyName The name of the property to be created.
     * @param string $propertyCode The unique code of the property.
     * @return void
     */
    private static function createOrderPropertyForAllPersonTypes($propertyName, $propertyCode)
    {
        // Получаем все типы плательщиков
        $personTypes = [];
        $dbPersonTypes = \CSalePersonType::GetList( ['SORT' => 'ASC'], [] );
        while ($personType = $dbPersonTypes->Fetch()) {
            $personTypes[] = $personType['ID'];
        }

        foreach ($personTypes as $personTypeId) {
            // Проверяем, существует ли уже свойство с таким кодом для данного типа плательщика
            $dbProps = \CSaleOrderProps::GetList(
                [],
                [ 'CODE' => $propertyCode, 'PERSON_TYPE_ID' => $personTypeId ]
            );

            if ($dbProps->Fetch())
                continue;

            // Создаем новую группу свойства
            $groupExists = false;
            $propsGroupId = null;
            $nameGroup = 'Генерация чека окончательного расчета';
            $groupCheck = \CSaleOrderPropsGroup::GetList( [],
                [ 'NAME' => $nameGroup, 'PERSON_TYPE_ID' => $personTypeId ]
            );

            if ($group = $groupCheck->Fetch()) {
                $groupExists = true;
                $propsGroupId = $group['ID'];
            }

            if (!$groupExists) {
                $groupFields = [
                    'NAME' => $nameGroup,
                    'PERSON_TYPE_ID' => $personTypeId,
                    'SORT' => 500
                ];
                $propsGroupId = \CSaleOrderPropsGroup::Add($groupFields);
            }

            // Создаем новое свойство
            $propertyFields = array(
                'NAME' => $propertyName,
                'TYPE' => 'STRING', // Тип свойства: STRING, NUMBER, FILE, etc.
                'REQUIRED' => 'N', // Обязательное: Y или N
                'DEFAULT_VALUE' => '',
                'SORT' => 500,
                'USER_PROPS' => 'Y', // Включить в свойства пользователя: Y или N
                'IS_LOCATION' => 'N',
                'IS_LOCATION4TAX' => 'N',
                'IS_EMAIL' => 'N',
                'IS_PROFILE_NAME' => 'N',
                'IS_PAYER' => 'N',
                'IS_FILTERED' => 'N',
                'CODE' => $propertyCode,
                'ACTIVE' => 'Y',
                'UTIL' => 'Y',
                'INPUT_FIELD_LOCATION' => '',
                'PROPS_GROUP_ID' => $propsGroupId, // ID группы свойств
                'SIZE1' => 0,
                'SIZE2' => 0,
                'DESCRIPTION' => '',
                'PERSON_TYPE_ID' => $personTypeId // Тип плательщика
            );
            \CSaleOrderProps::Add($propertyFields);
        }
    }

    /**
     * Triggers an event in case of an error in the Paykeeper API request.
     *
     * This method logs the error message and triggers the custom event 'onPrintCheckPaykeeper'.
     * Additional error handling can be performed in the event handler.
     *
     * @param string $errMessage The error message received from the Paykeeper API.
     * @param string $orderId The ID of the order that encountered the error.
     * @param string $nameEvent
     * @return void
     */
    private static function triggerPaykeeperEventError($errMessage, $orderId, $nameEvent) {
        $event = new Event('paykeeper', $nameEvent, [ $errMessage, $orderId ]);
        $event->send();

        foreach ($event->getResults() as $eventResult) {
            if ($eventResult->getType() == EventResult::ERROR && self::$logging) {
                $errorParams = $eventResult->getParameters();
                $log_params = [
                    'TITLE' => 'EVENT_ERROR',
                    'EVENT NAME' => $nameEvent,
                    'METHOD' => 'EVENT',
                    'DATA' => "ORDER ID $orderId MESSAGE $errMessage"
                ];
                if (!empty($errorParams['ERROR_MESSAGE'])) {
                    $log_params['TEXT ERROR'] = $errorParams['ERROR_MESSAGE'];
                }
                self::paykeeperLogger($log_params);
            }
        }
    }

    /**
     * Returns a successful event result.
     *
     * This method creates and returns an instance of `EventResult` with a success status.
     *
     * @return \Bitrix\Main\EventResult An event result object with a status of success.
     */
    private static function getSuccessEventResult() {
        return new EventResult(EventResult::SUCCESS);
    }

    /**
     * Retrieves the business value associated with a specified parameter for a given order.
     *
     * This method uses `BusinessValue::get()` to fetch the value of the provided parameter
     * based on the payment system ID and person type ID within the order.
     *
     * @param string $param The name of the business parameter to retrieve.
     * @param \Bitrix\Sale\Order $order The order object containing relevant payment system and person type IDs.
     *
     * @return mixed Returns the business value associated with the specified parameter for the payment system
     *               and person type in the order.
     */
    private static function getBusVal($param, $order)
    {
        return BusinessValue::get(
            $param,
            Service::PAY_SYSTEM_PREFIX . $order->getField('PAY_SYSTEM_ID'),
            $order->getField('PERSON_TYPE_ID')
        );
    }

    /**
     * Logs specified parameters to a log file with a timestamp.
     *
     * This method creates a log entry with the provided parameters and includes a timestamp.
     * If the log file exists and is less than 7 MB in size, the new log entry is prepended to the existing content.
     *
     * @param array $log_params An associative array of parameters to log, where each key is a parameter name and the value is its corresponding data.
     *
     * @return void
     */
    private static function paykeeperLogger($log_params)
    {
        $logText = '--------------START_' . date("Y-m-d_H:i:s") . '--------------' . "\n";
        foreach ($log_params as $key => $value) {
            $logText .= "$key: " . print_r($value,true) . "\n";
        }
        $logText .= "\n\n\n";
        $file = self::log_file;
        if (file_exists($file)) {
            $logSize = filesize($file);
            if ($logSize < (7 * 1024 * 1024)) {
                $logText .= file_get_contents($file);
            }
        }
        file_put_contents($file, $logText);
    }
}
